#define _GNU_SOURCE

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <bfd.h>

#ifdef WIN32
# include <windows.h>
# include <tlhelp32.h>
#else
# include <dlfcn.h>
// this cond is from ruby
# if defined(__hp9000s300) ||  (defined(__NetBSD__) && !defined(__ELF__)) || defined(__BORLANDC__) || (defined(__FreeBSD__) && !defined(__ELF__)) || (defined(__OpenBSD__) && !defined(__ELF__)) || defined(NeXT) || defined(__WATCOMC__) || defined(__APPLE__)
# include <mach-o/dyld.h>
#else
# include <link.h>
#endif
#endif

#include "reflection.h"

static int free_sym(void** ip, void* dummy) {
    reflection_sym_t* sym = (reflection_sym_t*)*ip;
    free(sym->name);
    free(sym);
    return 1;
}

int reflection_quit(reflection_t* ref) {
    int i;
    htab_traverse_noresize(ref->syms, free_sym, NULL);
    htab_delete(ref->syms);
    for (i = 0; i < ref->file_num; i++) {
        free(ref->files[i]);
    }
    free(ref->files);
    free(ref);
    return 0;
}

static int hash_string_eq(const void *s1_p, const void *s2_p) {
    const reflection_sym_t* s1 = (const reflection_sym_t*)s1_p;
    const char* s2 = (const char*)s2_p;
    return strcmp(s1->name, s2) == 0;
}

static char* mstrcpy(const char* s) {
    char* ret = (char*)malloc(sizeof(char)*(strlen(s)+1));
    strcpy(ret, s);
    return ret;
}

static void add_symbol_impl(reflection_t* ref, const char* name, int value) {
    reflection_sym_t** asymp;

    asymp = (reflection_sym_t**)
        htab_find_slot(ref->syms, name, INSERT);
    if (!*asymp) {
        *asymp = (reflection_sym_t*)malloc(sizeof(reflection_sym_t));
    }
    else {
        free((*asymp)->name);
    }
    (*asymp)->name = mstrcpy(name);
    (*asymp)->fp = value;

    ref->sym_num++;
//    printf("%08x %s\n", (*asymp)->fp, (*asymp)->name); 
}

static void add_symbol(reflection_t* ref, int offset, asymbol* asym) {
    const char* name;
    int value;

    name = bfd_asymbol_name(asym);
    value = bfd_asymbol_value(asym);

    if (name == NULL || strlen(name) == 0) return;

    if (value == 0) return;

#ifdef WIN32
    if (name[0] != '_') return;
    name++;
/*
    if (offset) {
        HMODULE cur_mod = (HMODULE)offset;
        offset = 0;
        value = GetProcAddress(cur_mod, name);
    }
*/
#endif

#ifdef __APPLE__
    if (name[0] != '_') return;
    name++;
    if (strchr(name, ':')) return;
#endif

    if (!isalpha((int)name[0])) return;

    if ((unsigned int)value < (unsigned int)offset) {
        value += offset;
    }

    add_symbol_impl(ref, name, value);
}

static int read_symbols(reflection_t* ref, const char* filename, int offset) {
    bfd* abfd;
    int i;
    int prev_symnum = ref->sym_num;
    int symnum;

    for (i = 0; i < ref->file_num; i++) {
        if (strcmp(ref->files[i], filename) == 0) return 0;
    }

    printf("reading %s(0x%x)... ", filename, offset);

    ref->files = (char**)realloc(ref->files, sizeof(char*)*(ref->file_num+1));
    ref->files[ref->file_num++] = mstrcpy(filename);

    abfd = bfd_openr(filename, NULL);
    if (abfd == NULL) {
        bfd_perror("bfd_openr");
        return 1;
    }
    if (!bfd_check_format(abfd, bfd_object)) {
        bfd_perror("bfd_check_format");
        return 1;
    }

#if 0
    {
        long storage_needed;
        asymbol **symbol_table;

        storage_needed = bfd_get_symtab_upper_bound(abfd);

        if (storage_needed < 0) {
            bfd_perror("bfd_get_symtab_upper_bound");
            goto error;
        }
        if (storage_needed == 0) {
            printf("no symbols\n");
            bfd_close(abfd);
            return 0;
        }
        symbol_table = (asymbol**)malloc(storage_needed);

        symnum = bfd_canonicalize_symtab(abfd, symbol_table);
        if (symnum < 0) {
            bfd_perror("bfd_canonicalize_symtab");
            goto error;
        }
        for (i = 0; i < symnum; i++) {
            asymbol* asym = symbol_table[i];
            add_symbol(ref, offset, asym);
/*
            reflection_sym_t** asymp;
            const char* name;

            name = bfd_asymbol_name(asym);
            asymp = (reflection_sym_t**)
                htab_find_slot(ref->syms, name, INSERT);
            *asymp = (reflection_sym_t*)malloc(sizeof(reflection_sym_t));
            (*asymp)->name = name;
            (*asymp)->fp = bfd_asymbol_value(asym);
            if ((unsigned int)(*asymp)->fp < (unsigned int)offset) {
                (*asymp)->fp += offset;
            }
//            printf("%08x %s\n", (*asymp)->fp, (*asymp)->name); 
*/
        }
        printf("load %d/%d symbols. (total %d)\n",
               ref->sym_num-prev_symnum, symnum, ref->sym_num);

        free(symbol_table);
    }
#else
    {
        void* minisyms;
        unsigned int size;
        char* p;
        asymbol* store;
        int dyn = 0;

        store = bfd_make_empty_symbol(abfd);

        symnum = bfd_read_minisymbols(abfd, dyn, &minisyms, &size);
#if defined(__linux__) || defined(__FreeBSD__)
        if (symnum == 0) {
            dyn = 1;
            symnum = bfd_read_minisymbols(abfd, dyn, &minisyms, &size);
        }
#endif
        if (symnum < 0) {
            bfd_perror("bfd_read_minisymbols");
            goto error;
        }

        p = (char*)minisyms;
        for (i = 0; i < symnum; i++) {
            asymbol* sym;
            sym = bfd_minisymbol_to_symbol(abfd, dyn, p, store);
            add_symbol(ref, offset, sym);
            p += size;
        }

        printf("load %d/%d symbols. (total %d)\n",
               ref->sym_num-prev_symnum, symnum, ref->sym_num);

        free(minisyms);
    }
#endif

    bfd_close(abfd);
    return 0;

error:
    bfd_close(abfd);
    return 1;
}

#if defined(WIN32)
static int load_arch(reflection_t* ref, MODULEENTRY32* me) {
    char buf[1024];
    FILE* fp;
    if (!GetModuleFileName(me->hModule, buf, 1024)) {
        perror("GetModuleFileName");
        return 1;
    }
    fp = fopen(buf, "rb");
    if (fp) {
        fclose(fp);
        return read_symbols(ref, buf, (int)me->modBaseAddr);
    }
    printf("%s: not found. Is it WINE?\n", buf);
    return 0;
}

static int init_arch(reflection_t* ref) {
    HMODULE h;
    HANDLE ss;
    MODULEENTRY32 me;

    printf("entering win32 initialize\n");

    ss = CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, 0);
    if (!ss) {
        perror("CreateToolhelp32Snapshot");
        return 1;
    }
    me.dwSize = 1024;
    if (!Module32First(ss, &me)) {
        perror("Module32First");
        return 1;
    }
    if (load_arch(ref, &me)) return 1;
    for (;;) {
        me.dwSize = 1024;
        if (!Module32Next(ss, &me)) {
            return 0;
        }
        if (load_arch(ref, &me)) return 1;
    }
    return 0;
}
#elif defined(__linux__)
static int regist_loaded(struct dl_phdr_info* info, size_t size, void* data) {
    if (!info->dlpi_name || !info->dlpi_name[0]) return 0;
    return read_symbols(data, info->dlpi_name, info->dlpi_addr);
}

static int init_arch(reflection_t* ref) {
    FILE* fp;
    if ((fp = fopen("/proc/self/exe", "rb"))) {
        fclose(fp);
        if (read_symbols(ref, "/proc/self/exe", 0)) return 1;
    }
    return dl_iterate_phdr(regist_loaded, ref);
}
#elif defined(__FreeBSD__) || (__sun__)
static int init_arch(reflection_t* ref) {
    struct link_map* lmap;
    if (dlinfo(RTLD_SELF, RTLD_DI_LINKMAP, &lmap) == 0) {
        while (lmap) {
            if (read_symbols(ref, lmap->l_name, lmap->l_addr)) return 1;
            lmap = lmap->l_next;
        }
        return 0;
    }
    else {
        perror("dlinfo");
        return 1;
    }
}
#elif defined(__APPLE__)
static int init_arch(reflection_t* ref) {
    int i;
    unsigned long cnt;

    cnt = _dyld_image_count();

    for (i = 0; i < cnt; i++) {
        const char* fname = _dyld_get_image_name(i);
        unsigned long vm_start = _dyld_get_image_vmaddr_slide(i);
        if (read_symbols(ref, fname, vm_start)) return 1;
    }

    return 0;
}
#else
#endif

reflection_t* reflection_init(const char* exe, int tbl_size) {
    reflection_t* ref;
    int ret;

    ref = (reflection_t*)malloc(sizeof(reflection_t));

    ref->files = NULL;
    ref->file_num = 0;
    ref->syms = htab_create(tbl_size, htab_hash_string, hash_string_eq, NULL);
    ref->sym_num = 0;

    bfd_init();

#if defined(WIN32) || defined(__linux__) || defined(__FreeBSD__) || defined(__APPLE__) || defined(__sun__)
    ret = init_arch(ref);
#else
    ret = read_symbols(ref, exe, 0);
/*    ret = reflection_dlopen(ref, exe, "main");*/
#endif

    if (ret) return NULL;
    else return ref;
}

#if defined(WIN32)
static int dlopen_arch(reflection_t* ref, const char* file, const char* func) {
    HMODULE h;
    int i = 0;
    char buf[256];

    if (!strchr(file, '.')) {
        strcpy(buf, file);
        strcat(buf, ".dll");
        file = buf;
    }

    h = LoadLibrary(file);
    if (!h) {
        perror("GetModuleHandle");
        return 1;
    }
    return init_arch(ref);
}
#else
static int dlopen_arch(reflection_t* ref, const char* file, const char* func) {
    void* dlhandle;
    char buf[256];

    if (!strchr(file, '/')) {
        if (strncmp(file, "lib", 3) != 0) {
            strcpy(buf, "lib");
            strcat(buf, file);
            file = buf;
        }
    }
    if (!strchr(file, '.')) {
        if (buf != file) strcpy(buf, file);
#ifdef __APPLE__
        strcat(buf, ".dylib");
#else
        strcat(buf, ".so");
#endif
        file = buf;
    }

    if (!(dlhandle = dlopen(file, RTLD_NOW))) {
        perror("cannot dlopen");
        return 1;
    }

#ifdef HAVE_DLINFO
    {
        struct link_map* lmap;

        if (dlinfo(dlhandle, RTLD_DI_LINKMAP, &lmap) == 0) {
            while (lmap) {
                if (read_symbols(ref, lmap->l_name, lmap->l_addr)) return 1;
                lmap = lmap->l_next;
            }
            return 0;
        }
        else {
            perror("dlinfo");
            return 1;
        }
    }
#else
    {
        void* fp;
        Dl_info info;
        if (!func) {
            perror("2nd argument is needed in this environment");
            return 1;
        }
        if (!(fp = dlsym(dlhandle, func))) {
            perror("dlsym");
            return 1;
        }
        if (!dladdr(fp, &info)) {
            perror("dladdr");
            return 1;
        }
        read_symbols(ref, info.dli_fname, info.dli_fbase);

        return 0;
    }
#endif
}
#endif

int reflection_dlopen(reflection_t* ref, const char* file, const char* func) {
    return dlopen_arch(ref, file, func);
}

static int reflection_dump_symbol_one(void** ip, void* dummy) {
    reflection_sym_t* sym = (reflection_sym_t*)*ip;
    printf("%08x %s\n", sym->fp, sym->name);
    return 1;
}

void reflection_dump_symbols(reflection_t* ref) {
    htab_traverse(ref->syms, reflection_dump_symbol_one, NULL);
}

#ifdef HAVE_FFI

#include <ffi.h>

ffi_type *get_ffi_type(reflection_type_t type) {
    switch (type) {
    case R_CHAR:
        return &ffi_type_sint8;
    case R_INT:
        return &ffi_type_sint32;
    case R_FLOAT:
        return &ffi_type_float;
    case R_PTR:
        return &ffi_type_pointer;
#ifdef HAVE_FFI
    case R_DOUBLE:
        return &ffi_type_double;
#endif
    default:
        return &ffi_type_void;
    }
}

#endif

int reflection_dispatch(
    reflection_t* ref, const char* name, reflection_type_t rtype,
    int arg_num, reflection_value_t* args, reflection_type_t* types,
    reflection_value_t* ret)
{
    reflection_sym_t** asym;
    void* fp = NULL;

    asym = (reflection_sym_t**)htab_find_slot(ref->syms, name, NO_INSERT);
    if (!asym) {
#ifndef WIN32
        fp = (void*)dlsym(RTLD_DEFAULT, name);
#endif
        if (fp == NULL) return 1;
    }
    else {
        fp = (void*)((*asym)->fp);
    }

#ifndef HAVE_FFI
# include "dispatch.c"
#else
    {
        ffi_cif cif;
        ffi_type* as[32];
        void* vs[32];
        ffi_arg rv;
        ffi_type* rt;
        int i;

        for (i = 0; i < arg_num; i++) {
            as[i] = get_ffi_type(types[i]);
            if (as[i] == &ffi_type_void) return 1;
            vs[i] = &args[i];
        }

        rt = get_ffi_type(rtype);
        if (ffi_prep_cif(&cif, FFI_DEFAULT_ABI, arg_num, rt, as) != FFI_OK) {
            return 1;
        }

        ffi_call(&cif, FFI_FN(fp), &rv, vs);

        if (rtype != R_VOID) {
            *ret = *(reflection_value_t*)&rv;
        }

        return 0;
    }
#endif
}

#ifdef TEST
int add(int a, int b) {
    return a+b;
}

int main(int argc, char *argv[]) {
    reflection_t* ref;

    ref = reflection_init(NULL, 10001);

    if (!ref) {
        printf("cannot init\n");
        exit(1);
    }

    {
        reflection_value_t ret;
        reflection_value_t args[2];
        reflection_type_t types[2];
        args[0].i = 2;
        args[1].i = 3;
        types[0] = R_INT;
        types[1] = R_INT;
        if (reflection_dispatch(ref, "add", R_INT, 2, args, types, &ret)) {
            printf("call add error\n");
            exit(1);
        }
        printf("%d\n", ret.i);
    }

    {
        if (reflection_dlopen(ref, "libm.so")) {
            printf("load libm error\n");
            exit(1);
        }
        reflection_value_t ret;
        reflection_value_t args[1];
        reflection_type_t types[1];
        args[0].f = 1.0;
        types[0] = R_FLOAT;
        if (reflection_dispatch(ref, "sinf", R_FLOAT, 1, args, types, &ret)) {
            printf("call sinf error\n");
            exit(1);
        }
        printf("%f\n", ret.f);  // about 0.84
    }

    return 0;
}
#endif
